# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import re
import six
import requests

from chandl import util
from chandl.model.post import Post

logger = logging.getLogger(__name__)


@six.python_2_unicode_compatible
class Thread:
    """
    Represents a 4Chan thread.
    """

    def __init__(self, board, id_, subject, title, slug, posts):
        """
        Initialise a new thread instance.

        :param board: The board this thread exists within.
        :param id_: The post id of the first post in this thread.
        :param subject: The thread subject, if set.
        :param title: A name for the thread, resolved from the subject, comment,
                      or thread id.
        :param slug: The thread's URL fragment.
        :param posts: A list of posts in this thread.
        """
        self.board = board
        self.id = id_
        self.subject = subject
        self.title = title
        self.slug = slug
        self.posts = posts

    @property
    def url(self):
        """
        Generate a link where this thread can be viewed.

        :return: The URL.
        """
        return 'https://boards.4chan.org/{0}/thread/{1}/{2}'.format(
            self.board, self.id, self.slug)

    @staticmethod
    def _find_subject(post):
        """
        Identifies the most appropriate name for a thread.

        :param post: The first post's json.
        :return: A string containing the thread's resolved subject.
        """

        # if the post has a subject, it's easy
        if 'sub' in post:
            return util.unescape_html(post['sub'])

        # return the first sentence of the post content
        comment = util.unescape_html(post['com'])
        result = re.match(r'([^.:;?]+)', comment)
        if result:
            return result.group(1)

        # fall back to the post id
        return str(post['no'])

    @staticmethod
    def parse_json(board, json_):
        """
        Create a thread instance from JSON returned by the 4Chan API.

        :param board: The board that was requested.
        :param json_: The thread's parsed JSON as a dictionary.
        :return: The created thread instance.
        """
        if 'posts' not in json_ or not json_['posts']:
            raise ValueError('Thread does not contain any posts')

        first = json_['posts'][0]

        return Thread(board,
                      first['no'],
                      util.unescape_html(first['sub'])
                      if 'sub' in first else None,
                      Thread._find_subject(first),
                      first['semantic_url'],
                      [Post.parse_json(board, post)
                       for post in json_['posts']])

    @staticmethod
    def from_url(url, session=None):
        """
        Construct a thread instance from its URL.

        :param url: The URL of the thread to retrieve.
        :param session: The requests session to use to send the request.
        :return: The created thread instance.
        :raises IOError: If the thread could not be retrieved from 4chan.
        """
        # extract the board and thread ids
        result = re.search('boards\.4chan\.org/([a-z]+)/thread/([0-9]+)', url)
        if not result:
            raise ValueError('Invalid thread URL: {0}'.format(url))

        # construct a session if necessary
        if not session:
            session = util.create_session()

        # determine the URL
        api_url = 'https://a.4cdn.org/{0}/thread/{1}.json'.format(
            result.group(1), result.group(2))

        # download the JSON
        logger.debug('Retrieving JSON from %s', api_url)
        response = session.get(api_url)
        if response.status_code != requests.codes.ok:
            raise IOError('Request to 4chan failed with status code {0}'.format(
                response.status_code))
        try:
            return Thread.parse_json(result.group(1), response.json())
        except ValueError as e:
            raise IOError('Error parsing 4chan response: {0}'.format(e))

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and other.__dict__ == self.__dict__

    def __str__(self):
        return 'Thread({0}, {1})'.format(self.id, self.board)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import os
import binascii
import base64
import six
import requests
import shutil

from chandl import util


TYPE_VIDEO = ['webm', 'gif']
TYPE_IMAGE = ['jpg', 'png']

# passable via the CLI's --filter option
_FILTERS = {
    'videos': TYPE_VIDEO,
    'images': TYPE_IMAGE
}

logger = logging.getLogger(__name__)


def expand_filters(filters):
    """
    Expand a list of file filters passed on the command line. Each item could be
    a csv or single file extension, or key in the _FILTERS dict above, in which
    case it needs to be expanded into file extensions. Examples:

        ['jpg', 'webm'] -> jpg, webm
        ['jpg, webm', 'png'] -> jpg, webm, png
        ['webm,images'] -> webm, jpg, png

    :param filters: The list of filters from argparse.
    :return: A set of file extensions.
    """
    extensions = set()

    for arg in filters:  # "videos, jpg"
        for type_ in [s.strip() for s in arg.split(',')]:  # "videos"|"jpg"
            if type_ in _FILTERS:  # "videos"
                for ext in _FILTERS[type_]:
                    # return each extension in that category
                    extensions.add(ext)
            else:  # "jpg"
                # it's a file extension - just return it
                extensions.add(type_)

    return extensions


@six.python_2_unicode_compatible
class File:
    """
    Represents a media file attached to a post.
    """

    def __init__(self, id_, board, name, extension, size, width, height, md5):
        """
        Initialise a new file instance.

        :param id_: A unique identifier for the file: it's UNIX timestamp
                    prepended to the number of milliseconds.
        :param board: The board this file was posted in, e.g. 'wg'.
        :param name: The original name of the file.
        :param extension: The file's extension, without leading dot.
        :param size: The size of the file in bytes.
        :param width: The width of the media in pixels.
        :param height: The height of the media in pixels.
        :param md5: The MD5 hash of the file as a 32-character lowercase string.
        """
        self.id = id_
        self.board = board
        self.name = name
        self.extension = extension
        self.size = size
        self.width = width
        self.height = height
        self.md5 = md5

    @property
    def filename(self):
        """
        Get the name of this file as it appears on the 4chan website.

        :return: This file's id with extension.
        """
        return six.u(str(self.id)) + '.' + self.extension

    @property
    def url(self):
        """
        Retrieve the URL where this file can be downloaded.

        :return: The URL.
        """
        return 'https://i.4cdn.org/{0}/{1}'.format(self.board, self.filename)

    def save_to(self, directory, name, verify=True, session=None):
        """
        Download and save this file.

        :param directory: The directory to save this file within.
        :param name: The file name to save under.
        :param verify: Whether to verify the file's checksum once it is written.
                       Defaults to true.
        :param session: The requests session to use for this download. If
                        omitted, a new session will be used.
        :raise IOError: If the file could not be downloaded, written, or if
                        `verify` was enabled and its checksum did not match the
                        one reported by 4chan.
        """
        destination = os.path.join(directory, name)

        if os.path.isfile(destination) and \
                util.md5_file(destination) == self.md5:
            logger.debug('%s already exists; skipping download', self)
            return

        logger.debug('Downloading %s', self)
        if not session:
            session = util.create_session()
        response = session.get(self.url, stream=True)
        if response.status_code != requests.codes.ok:
            raise IOError('File failed to download with status {0}'.format(
                response.status_code))

        with open(destination, 'wb') as handle:
            shutil.copyfileobj(response.raw, handle)

        if verify and util.md5_file(destination) != self.md5:
            raise IOError('Verify failed: checksum mismatch')

    @staticmethod
    def parse_json(board, json):
        """
        Create a file object from 4Chan's API format.

        :param board: The name of the board this file was uploaded to, e.g.
                      'wg'.
        :param json: The JSON of the post containing this file.
        :return: The created file instance.
        :raises ValueError: If the post does not contain an image.
        """

        def unpack_hash(encoded):
            """
            Unpacks a packed, base64 encoded MD5 checksum.
            :param encoded: The unicode representation of the encoded checksum.
            :return: The unicode representation of the unpacked, 32-character
                     checksum.
            """
            return binascii.hexlify(
                base64.b64decode(encoded.encode('ascii'))).decode('ascii')

        if 'filename' not in json:
            raise ValueError('Post does not contain an image')

        return File(json['tim'], board, util.unescape_html(json['filename']),
                    json['ext'][1:], json['fsize'], json['w'], json['h'],
                    unpack_hash(json['md5']).rstrip())

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
               other.__dict__ == self.__dict__

    def __str__(self):
        return 'File({0}, {1}.{2}, {3}, {4}x{5})'.format(
            self.id, self.name, self.extension, util.bytes_fmt(self.size),
            self.width, self.height)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import unittest
import sys
import os
import six
import logging

import chandl
from chandl import util


class TestBytesFmt(unittest.TestCase):

    _TRIALS = {
        32242: '31.5 KiB',
        0: '0.0 B',
        -1: '1.0 B',
        1024: '1.0 KiB',
        1024 * 1024: '1.0 MiB',
        1024 * 2 ** 70: '1.0 YiB'
    }

    def test_none(self):
        with self.assertRaises(ValueError):
            util.bytes_fmt(None)

    def test_trials(self):
        for number, formatted in six.iteritems(self._TRIALS):
            self.assertEqual(util.bytes_fmt(number), formatted)


class TestDecodeCliArg(unittest.TestCase):

    _ARG_VALUE = 'test_arg_value'

    def test_empty(self):
        with self.assertRaises(ValueError):
            util.decode_cli_arg(None)

    @unittest.skipUnless(sys.version_info.major == 2,
                         'Only applies to Python 2')
    def test_valid_2(self):
        self.assertEqual(
            util.decode_cli_arg(
                self._ARG_VALUE.encode(sys.getfilesystemencoding())),
            self._ARG_VALUE)

    @unittest.skipUnless(sys.version_info.major == 3,
                         'Only applies to Python 3')
    def test_valid_3(self):
        self.assertEqual(util.decode_cli_arg(self._ARG_VALUE), self._ARG_VALUE)


class TestExpandCliArgs(unittest.TestCase):

    def test_empty(self):
        self.assertEqual(util.expand_cli_args([]),
                         set())

    def test_strip(self):
        self.assertEqual(util.expand_cli_args(['  abc,  def ']),
                         {'abc', 'def'})

    def test_csv(self):
        self.assertEqual(util.expand_cli_args(['ghi,jkl , klm']),
                         {'ghi', 'jkl', 'klm'})

    def test_mixed(self):
        self.assertEqual(util.expand_cli_args(['nop', 'qrs,tuv']),
                         {'nop', 'qrs', 'tuv'})

    def test_duplicates(self):
        self.assertEqual(util.expand_cli_args(['wxyz', 'wxyz,wxyz']),
                         {'wxyz'})


class TestMakeFilename(unittest.TestCase):

    _NO_CHANGE = ['file_name.jpg',
                  'File name2.png',
                  '124_nAme22.png.abc']

    _MODIFIED = {
        'my%file"name().png': 'myfilename.png',
        '$*$*.webm': '.webm'
    }

    _UNICODE = {
        'Sîne klâwen durh': 'Sine klawen durh',
        'Τη γλώσσα μου έδωσαν': 'Te glossa mou edosan',
        'ღმერთსი შემვედრე': 'g\'mertsi shemvedre'
    }

    def test_none(self):
        with self.assertRaises(ValueError):
            util.make_filename(None)

    def test_empty(self):
        with self.assertRaises(ValueError):
            print(util.make_filename('$$$'))

    def test_no_change(self):
        for name in self._NO_CHANGE:
            self.assertEqual(util.make_filename(name), name)

    def test_modified(self):
        for original, mutated in six.iteritems(self._MODIFIED):
            self.assertEqual(util.make_filename(original), mutated)

    def test_unicode(self):
        for original, mutated in six.iteritems(self._UNICODE):
            self.assertEqual(util.make_filename(original), mutated)


class TestMd5File(unittest.TestCase):

    _TESTS_DIRECTORY = os.path.join(os.path.dirname(__file__), 'md5_tests')

    def test_none(self):
        with self.assertRaises(ValueError):
            util.md5_file(None)

    def test_not_file(self):
        with self.assertRaises(IOError):  # no FileNotFoundError in 2.7
            util.md5_file('does_not_exist')

    def test_valid_hashes(self):
        for filename in os.listdir(self._TESTS_DIRECTORY):
            path = os.path.join(self._TESTS_DIRECTORY, filename)
            self.assertEqual(util.md5_file(path), os.path.basename(path))


class TestLogLevelFromVerbosity(unittest.TestCase):

    def test_warning(self):
        self.assertEqual(util.log_level_from_vebosity(0), logging.WARNING)

    def test_info(self):
        self.assertEqual(util.log_level_from_vebosity(1), logging.INFO)

    def test_debug(self):
        self.assertEqual(util.log_level_from_vebosity(2), logging.DEBUG)
        self.assertEqual(util.log_level_from_vebosity(None), logging.DEBUG)


class TestCreateSession(unittest.TestCase):

    def test_user_agent(self):
        session = util.create_session()
        self.assertIn('User-Agent', session.headers)
        self.assertEqual(session.headers['User-Agent'],
                         'chandl/' + chandl.__version__)


class TestUnescapeHtml(unittest.TestCase):

    def test_all_encoded(self):
        self.assertEqual(
            util.unescape_html('&#x46;&#x6F;&#x6F;&#x20;&#xA9;&#x20;&#x62;&#x61'
                               ';&#x72;&#x20;&#x1D306;&#x20;&#x62;&#x61;&#x7A;&'
                               '#x20;&#x2603;&#x20;&#x71;&#x75;&#x78;'),
            'Foo © bar 𝌆 baz ☃ qux')

    def test_exotic_characters(self):
        self.assertEqual(
            util.unescape_html('Foo &#xA9; bar &#x1D306; baz &#x2603; qux'),
            'Foo © bar 𝌆 baz ☃ qux')

    def test_named_reference(self):
        self.assertEqual(util.unescape_html('&copy;'), '©')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# N.B. Be careful what is imported in this file; it is included in setup.py
#      so dependencies are not necessarily installed.

__version__ = '0.3.0'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import unittest
import datetime
import json
import pytz
from httmock import all_requests, response, HTTMock

from chandl.model.file import File
from chandl.model.post import Post
from chandl.model.thread import Thread


class TestThread(unittest.TestCase):

    _THREAD_JSON = {
        'posts': [
            {
                'bumplimit': 0,
                'com': 'No more, no less',
                'ext': '.jpg',
                'filename': 'IMG_5433',
                'fsize': 35649,
                'h': 768,
                'imagelimit': 0,
                'images': 12,
                'md5': '56yiIJJgznhqupf3uu2xuA==',
                'name': 'Anonymous',
                'no': 6840627,
                'now': '02/02/17(Thu)01:25:15',
                'replies': 20,
                'resto': 0,
                'semantic_url': 'monty-python-papes',
                'sub': 'Monty Python Papes',
                'tim': 1486016715106,
                'time': 1486016715,
                'tn_h': 140,
                'tn_w': 250,
                'unique_ips': 18,
                'w': 1366
            },
            {
                'com': 'I love that... wish there WERE more monty python papes',
                'name': 'Anonymous',
                'no': 6841819,
                'now': '02/03/17(Fri)03:43:48',
                'resto': 6840627,
                'time': 1486111428
            },
            {
                'ext': '.jpg',
                'filename': 'stretched-32086',
                'fsize': 587095,
                'h': 1080,
                'md5': '3tdY3WI0UqMC+Ia+fLuyug==',
                'name': 'Anonymous',
                'no': 6842016,
                'now': '02/03/17(Fri)09:53:15',
                'resto': 6840627,
                'tim': 1486133595824,
                'time': 1486133595,
                'tn_h': 70,
                'tn_w': 125,
                'w': 1920
            },
            {
                'com': '<a href="#p6840627" class="quotelink">&gt;&gt;6840627</'
                       'a><br>All I have. Thanks for the sparrow one!',
                'ext': '.jpg',
                'filename': '1475710924523',
                'fsize': 74606,
                'h': 1080,
                'md5': 'DcWYVYrdxLwagDoS6D4e/w==',
                'name': 'Anonymous',
                'no': 6842026,
                'now': '02/03/17(Fri)10:06:10',
                'resto': 6840627,
                'tim': 1486134370795,
                'time': 1486134370,
                'tn_h': 70,
                'tn_w': 125,
                'w': 1920
            }
        ]
    }
    _BOARD = 'wg'
    _ID = 6840627
    _SUBJECT = 'Monty Python Papes'
    _TITLE = _SUBJECT
    _SLUG = 'monty-python-papes'
    _POSTS = [
        Post(_BOARD, 6840627,
             datetime.datetime(2017, 2, 2, 6, 25, 15, tzinfo=pytz.utc),
             'No more, no less',
             File(1486016715106, _BOARD, 'IMG_5433', 'jpg', 35649, 1366,
                  768, 'e7aca2209260ce786aba97f7baedb1b8')),
        Post(_BOARD, 6841819,
             datetime.datetime(2017, 2, 3, 8, 43, 48, tzinfo=pytz.utc),
             'I love that... wish there WERE more monty python papes'),
        Post(_BOARD, 6842016,
             datetime.datetime(2017, 2, 3, 14, 53, 15, tzinfo=pytz.utc),
             file_=File(1486133595824, _BOARD, 'stretched-32086', 'jpg', 587095,
                        1920, 1080, 'ded758dd623452a302f886be7cbbb2ba')),
        Post(_BOARD, 6842026,
             datetime.datetime(2017, 2, 3, 15, 6, 10, tzinfo=pytz.utc),
             '<a href="#p6840627" class="quotelink">>>6840627</a><br>All I have'
             '. Thanks for the sparrow one!',
             File(1486134370795, _BOARD, '1475710924523', 'jpg', 74606, 1920,
                  1080, '0dc598558addc4bc1a803a12e83e1eff'))
    ]

    _VALID_URL = 'http://boards.4chan.org/wg/thread/6847183'

    @classmethod
    def setUpClass(cls):
        cls._thread = Thread(cls._BOARD, cls._ID, cls._SUBJECT, cls._TITLE,
                             cls._SLUG, cls._POSTS)

    def test_url(self):
        self.assertEqual(
            self._thread.url,
            'https://boards.4chan.org/{0}/thread/{1}/{2}'.format(self._BOARD,
                                                                 self._ID,
                                                                 self._SLUG))

    def test_find_subject_subject_raw(self):
        self.assertEqual(
            Thread._find_subject({
                'sub': 'Monty Python Papes'
            }),
            'Monty Python Papes')

    def test_find_subject_subject_escaped(self):
        self.assertEqual(
            Thread._find_subject({
                'sub': '&gt;&gt; abcd &copy;&#163;'
            }),
            '>> abcd ©£')

    def test_find_subject_comment_single_sentence(self):
        self.assertEqual(
            Thread._find_subject({
                'com': 'this is the first - sentence.'
            }),
            'this is the first - sentence')

    def test_find_subject_comment_multiple_sentence(self):
        self.assertEqual(
            Thread._find_subject({
                'com': 'this is the first - sentence. here\'s another'
            }),
            'this is the first - sentence')

    def test_find_subject_fallback(self):
        self.assertEqual(
            Thread._find_subject({
                'no': 6840627,
                'com': '... abcd efgh'
            }),
            '6840627')

    def test_parse_json(self):
        self.assertEqual(Thread.parse_json(self._BOARD, self._THREAD_JSON),
                         self._thread)

    def test_parse_json_no_posts(self):
        with self.assertRaises(ValueError):
            Thread.parse_json(self._BOARD, {'posts': []})

    def test_from_url_invalid_url(self):
        with self.assertRaises(ValueError):
            Thread.from_url('http://boards.4chan.org/thread/abcd')

    def test_from_url_404(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            return response(404)

        with HTTMock(response_content), self.assertRaises(IOError):
            Thread.from_url(self._VALID_URL)

    def test_from_url_503(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            return response(403)

        with HTTMock(response_content), self.assertRaises(IOError):
            Thread.from_url(self._VALID_URL)

    def test_from_url_invalid_json(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            return response(content='invalid json here')

        with HTTMock(response_content), self.assertRaises(IOError):
            Thread.from_url(self._VALID_URL)

    def test_from_url_malformed_response(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            return response(content='{}')

        with HTTMock(response_content), self.assertRaises(IOError):
            Thread.from_url(self._VALID_URL)

    def test_from_url(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            return response(content=json.dumps(self._THREAD_JSON))

        with HTTMock(response_content):
            self.assertEqual(Thread.from_url(self._VALID_URL),
                             self._thread)

    def test_str(self):
        self.assertEqual(str(self._thread),
                         'Thread({0}, {1})'.format(self._ID, self._BOARD))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import sys
import hashlib
import unidecode
import six
import requests

import chandl


def bytes_fmt(num, suffix='B'):
    """
    Turn a number of bytes into a more friendly representation, e.g. 2.5MiB.

    :param num: The number of bytes to convert.
    :param suffix: The unit suffix (defaults to 'B').
    :return: The human-readable equivalent of the input size.
    :raises ValueError: If num is not an integer.
    """
    if not isinstance(num, six.integer_types):
        raise ValueError('Byte count must be an integral type')

    num = abs(num)
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if num < 1024.0:
            return '{0:.1f} {1}{2}'.format(num, unit, suffix)
        num /= 1024.0
    return '{0:.1f} {1}{2}'.format(num, 'Yi', suffix)


def decode_cli_arg(arg):
    """
    Turn a bytestring provided by `argparse` into unicode.

    :param arg: The bytestring to decode.
    :return: The argument as a unicode object.
    :raises ValueError: If arg is None.
    """
    if arg is None:
        raise ValueError('Argument cannot be None')

    if sys.version_info.major == 3:
        # already decoded
        return arg

    return arg.decode(sys.getfilesystemencoding())


def expand_cli_args(args):
    """
    Expand a list of possibly comma separated arguments, removing duplicates.

    :param args: The list of arguments to expand.
    :return: The set of unique arguments.
    """
    items = set()
    for arg in args:  # "a.jpg,b.png"
        for arg_ in [n.strip() for n in arg.split(',')]:  # "a.jpg"|"b.jpg"
            items.add(arg_)
    return items


def make_filename(string):
    """
    Turn a string into something that can be safely used as a file or directory
    name.

    :param string: The string to convert.
    :return: The sanitised string.
    :raises ValueError: If string is None.
    """
    if string is None:
        raise ValueError('String cannot be None')

    safe = [' ', '.', '_', '-', '\'']
    joined = ''.join([c for c in unidecode.unidecode(string)
                      if c.isalnum() or c in safe]).strip()
    if not joined:
        raise ValueError('Filename would be empty')
    return joined


def md5_file(path):
    """
    Get the MD5 hash of a file.

    :param path: The path of the file.
    :return: The 32-character long lowercase hex representation of the
             checksum.
    :raises ValueError: If path is invalid.
    """
    if not path:
        raise ValueError('Path cannot be empty or None')

    hash_ = hashlib.md5()
    with open(path, 'rb') as fd:
        for chunk in iter(lambda: fd.read(4096), b''):
            hash_.update(chunk)
    return hash_.hexdigest()


def log_level_from_vebosity(verbosity):
    """
    Get the `logging` module log level from a verbosity.

    :param verbosity: The number of times the `-v` option was specified.
    :return: The corresponding log level.
    """
    if verbosity == 0:
        return logging.WARNING
    if verbosity == 1:
        return logging.INFO
    return logging.DEBUG


def create_session():
    """
    Create a requests session for issuing HTTP requests to 4chan.

    :return: The created session.
    """
    headers = requests.utils.default_headers()
    headers.update({
        'User-Agent': 'chandl/' + chandl.__version__
    })

    session = requests.Session()
    session.headers = headers
    return session


def unescape_html(html_):
    """
    Replace HTML entities (e.g. `&pound;`) in a string.

    :param html_: The escaped HTML.
    :return: The input string with entities replaces.
    """

    # http://stackoverflow.com/a/2360639

    if sys.version_info.major == 2:  # 2.7
        # noinspection PyUnresolvedReferences,PyCompatibility
        from HTMLParser import HTMLParser
        return HTMLParser().unescape(html_)

    if sys.version_info.minor == 3:  # 3.3
        # noinspection PyCompatibility
        from html.parser import HTMLParser
        # noinspection PyDeprecation
        return HTMLParser().unescape(html_)

    # 3.4+
    # noinspection PyCompatibility
    import html
    return html.unescape(html_)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import multiprocessing
import time
import signal
import functools
import threading
import collections
import contextlib
import six
import datetime
import os
import requests
from progress.bar import Bar

from chandl import util


logger = logging.getLogger(__name__)

_interrupted = False


# noinspection PyUnusedLocal
def _handle_sigint(number, frame):
    """
    Called when we receive a SIGINT.

    :param number: The signal number (2 in this case).
    :param frame: The stack frame.
    """
    global _interrupted
    _interrupted = True


@contextlib.contextmanager
def _redirect_sigint():
    """
    For the duration of this function, a SIGINT will call _handle_sigint()
    rather than the normal Python routine.
    """
    original_handler = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, _handle_sigint)
    try:
        yield
    except:
        raise
    finally:
        signal.signal(signal.SIGINT, original_handler)


@six.python_2_unicode_compatible
class DownloadResult:
    """
    Represents the outcome of a thread download.
    """

    @staticmethod
    def _posts_size(posts):
        """
        Find the total size of the files in all posts in a list.

        :param posts: The posts whose files to count.
        :return: The total size of the files in the posts in bytes.
        """
        return sum([post_.file.size for post_ in posts if post_.has_file])

    def __init__(self, completed_jobs, remaining_jobs, elapsed):
        """
        Initialise a new download result.

        :param completed_jobs: A list of posts that downloaded successfully.
        :param remaining_jobs: A list of posts that were not downloaded.
        :param elapsed: A timedelta representing the duration of the download.
        """
        self.downloaded_bytes = self._posts_size(completed_jobs)
        self.remaining_bytes = self._posts_size(remaining_jobs)
        self.total_bytes = self.downloaded_bytes + self.remaining_bytes

        self.completed_job_count = len(completed_jobs)
        self.remaining_job_count = len(remaining_jobs)
        self.total_jobs = self.completed_job_count + self.remaining_job_count

        self.completed_jobs = completed_jobs
        self.remaining_jobs = remaining_jobs
        self.elapsed = elapsed

    def __str__(self):
        """
        Get a formatted representation of this download's outcome.

        :return: The download statistics as a human-readable string.
        """
        string = '{0}/{1} jobs completed, {2} remaining{3}'.format(
            self.completed_job_count,
            self.total_jobs,
            self.remaining_job_count,
            os.linesep)
        string += '{0}/{1} downloaded, {2} remaining{3}'.format(
            util.bytes_fmt(self.downloaded_bytes),
            util.bytes_fmt(self.total_bytes),
            util.bytes_fmt(self.remaining_bytes),
            os.linesep)
        string += 'Duration: {0:0.3f} seconds ({1}/s)'.format(
            self.elapsed.total_seconds(),
            util.bytes_fmt(int(self.downloaded_bytes //
                               self.elapsed.total_seconds())))
        return string


class Downloader:
    """
    A basic thread pool implementation to download multiple files
    simultaneously.
    """

    def __init__(self, directory, name_fmt, parallelism=4):
        """
        Initialise a new downloader instance. Instances should not be reused.

        :param directory: The directory to save files in.
        :param name_fmt: A format specifier for file names.
        :param parallelism: The maximum number of threads to use to download
                            files per CPU. E.g. parallelism of 4 on a quad core
                            results in 16 threads.
        """
        self._directory = directory
        self._name_fmt = name_fmt
        self._threads = multiprocessing.cpu_count() * parallelism
        self._queue = collections.deque()

        self._lock = threading.Lock()

        self._completed_jobs = []

    @staticmethod
    def runner(self):
        """
        A single thread's execution.

        :param self: The downloader instance the thread belongs to.
        """
        session = requests.Session()
        while not _interrupted:
            try:
                post_ = self._queue.popleft()
                Downloader.handle(self, post_, session)
            except IndexError:
                # no items left to process - let function return
                break

    @staticmethod
    def handle(self, post_, session):
        """
        Downloads the file in a post.

        :param self: The downloader context.
        :param post_: The post to download.
        :param session: The requests session to use for the download.
        """
        try:
            name = self._name_fmt.format(**post_.__dict__)
            post_.file.save_to(self._directory, name, session=session)
            with self._lock:
                self._completed_jobs.append(post_)
        except IOError as e:
            logger.exception('Failed to write %s: %s', post_.file, e.message)

    def _queue_all(self, posts):
        """
        Add all posts in an iterable to the download queue.

        :param posts: The posts to add.
        """
        for post_ in posts:
            self._queue.append(post_)

    def download(self, posts, show_progress=False):
        """
        Download the files contained within a list of posts.

        :param posts: An iterable containing the posts to download. They will
                      be downloaded in order.
        :param show_progress: Whether to print a progress bar that updates as
                              the thread is downloading. Defaults to false.
        """
        global _interrupted
        _interrupted = False

        # populate the queue
        self._queue_all(posts)
        job_count = len(self._queue)

        # don't launch more threads than files
        threads = min(self._threads, len(self._queue))
        logger.debug('Will use %d threads for downloading', threads)

        # launch threads
        thread_pool = []
        target = functools.partial(Downloader.runner, self)
        start = datetime.datetime.now()
        for _ in range(threads):
            thread = threading.Thread(target=target)
            thread.start()
            thread_pool.append(thread)
        logger.debug('All threads launched')

        if show_progress:
            bar = Bar('Downloading',
                      max=job_count,
                      suffix='%(index)d/%(max)d - '
                             '%(elapsed_td)s elapsed, %(eta_td)s remaining')
            with _redirect_sigint():
                while len(self._completed_jobs) < job_count and \
                        not _interrupted:
                    bar.goto(len(self._completed_jobs))
                    time.sleep(.5)

        # wait for all threads to finish
        for i in range(threads):
            thread_pool[i].join()
        finish = datetime.datetime.now()

        if show_progress:
            # complete the progress bar if the queue is empty
            if not self._queue:
                # noinspection PyUnboundLocalVariable
                bar.goto(job_count)
            bar.finish()

        remaining_jobs = list(self._queue)
        return DownloadResult(self._completed_jobs,
                              remaining_jobs,
                              finish - start)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from unittest import TestCase
import datetime
import pytz

from chandl import util
from chandl.model.file import File
from chandl.model.post import Post


class TestPost(TestCase):

    BOARD = 'wg'

    POST_JSON = {
        'no': 6849229,
        'now': '02/11/17(Sat)21:33:46',
        'name': 'Anonymous',
        'sub': 'Financial backgrounds',
        'com': '<a href=\"#p1978934\" class=\"quotelink\">&gt;&gt;1978934</a>'
               '<br>4/?',
        'filename': 'wall_2',
        'ext': '.jpg',
        'w': 1200,
        'h': 934,
        'tn_w': 250,
        'tn_h': 194,
        'tim': 1486866826992,
        'time': 1486866826,
        'md5': 'iKLZQLYdEGgYi/xuNnd9nQ==',
        'fsize': 270555,
        'resto': 1978004,
        'bumplimit': 0,
        'imagelimit': 0,
        'semantic_url': 'financial-backgrounds',
        'replies': 2,
        'images': 2,
        'unique_ips': 1
    }
    POST_JSON_FILE_MD5 = '88a2d940b61d1068188bfc6e36777d9d'
    POST_JSON_FILE_URL = 'https://i.4cdn.org/wg/1486866826992.jpg'

    POST = Post(BOARD, POST_JSON['no'],
                datetime.datetime(2017, 2, 12, 2, 33, 46, tzinfo=pytz.utc),
                '<a href=\"#p1978934\" class=\"quotelink\">>>1978934</a>'
                '<br>4/?', File.parse_json(BOARD, POST_JSON))

    _POST_NO_BODY_JSON = {
        'no': 1978719,
        'now': '11/22/16(Tue)11:22:20',
        'name': 'Anonymous',
        'filename': '0guzMVA',
        'ext': '.jpg',
        'w': 1920,
        'h': 1200,
        'tn_w': 125,
        'tn_h': 78,
        'tim': 1479831740360,
        'time': 1479831740,
        'md5': 'cpoZmr+1xg58LzT2DIK2ZQ==',
        'fsize': 128012,
        'resto': 1978004
    }

    _POST_NO_BODY = Post(BOARD, _POST_NO_BODY_JSON['no'],
                         datetime.datetime(2016, 11, 22, 16, 22, 20,
                                           tzinfo=pytz.utc),
                         file_=File.parse_json(BOARD, _POST_NO_BODY_JSON))

    POST_NO_FILE_JSON = {
        'no': 1978945,
        'now': '11/23/16(Wed)04:52:24',
        'name': 'Anonymous',
        'com': '<a href=\"#p1978196\" class=\"quotelink\">&gt;&gt;1978196</a>'
               '<br>This is some HTML.',
        'time': 1479894744,
        'resto': 1978004
    }

    _POST_NO_FILE = Post(BOARD, POST_NO_FILE_JSON['no'],
                         datetime.datetime(2016, 11, 23, 9, 52, 24,
                                           tzinfo=pytz.utc),
                         '<a href=\"#p1978196\" class=\"quotelink\">>>1978196</'
                         'a><br>This is some HTML.')

    def test_board(self):
        self.assertEqual(self.POST.board, self.BOARD)

    def test_id(self):
        self.assertEqual(self.POST.id, self.POST_JSON['no'])

    def test_timestamp(self):
        self.assertEqual(self.POST.timestamp,
                         datetime.datetime.utcfromtimestamp(
                             self.POST_JSON['time']).replace(tzinfo=pytz.utc))

    def test_body(self):
        self.assertEqual(self.POST.body,
                         util.unescape_html(self.POST_JSON['com']))

    def test_no_body(self):
        self.assertIsNone(self._POST_NO_BODY.body)

    def test_file(self):
        self.assertEqual(self.POST.id, self.POST_JSON['no'])

    def test_no_file(self):
        self.assertIsNone(self._POST_NO_FILE.file)

    def test_has_file_true(self):
        self.assertTrue(self.POST.has_file)

    def test_has_file_false(self):
        self.assertFalse(self._POST_NO_FILE.has_file)

    def test_parse_json(self):
        self.assertEqual(Post.parse_json(self.BOARD, self.POST_JSON),
                         self.POST)

    def test_str(self):
        self.assertEqual(
            str(self.POST),
            'Post({0}, {1}, {2})'.format(
                self.POST_JSON['no'],
                str(datetime.datetime.utcfromtimestamp(
                    self.POST_JSON['time']).replace(tzinfo=pytz.utc)),
                str(self.POST.file)))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from datetime import datetime
import six
import pytz

from chandl import util
from chandl.model.file import File


@six.python_2_unicode_compatible
class Post:
    """
    Represents a 4Chan post.
    """

    def __init__(self, board, id_, timestamp, body=None, file_=None):
        """
        Initialise a new post instance.

        :param board: The board this post was submitted to.
        :param id_: The post number.
        :param timestamp: When this post was submitted.
        :param body: The body of this post, if one exists.
        :param file_: This post's file, if one exists.
        """
        self.board = board
        self.id = id_
        self.timestamp = timestamp
        self.body = body
        self.file = file_

    @property
    def has_file(self):
        """
        Find whether this post has a file.

        :return: True if it does, false otherwise.
        """
        return self.file is not None

    @staticmethod
    def parse_json(board, json):
        """
        Create a post instance from JSON returned by the 4Chan API.

        :param board: The board that was requested.
        :param json: The post's parsed JSON as a dictionary.
        :return: The created post instance.
        """
        file_ = File.parse_json(board, json) if 'tim' in json else None
        comment = util.unescape_html(json['com']) if 'com' in json else None
        return Post(board, json['no'],
                    datetime.utcfromtimestamp(
                        json['time']).replace(tzinfo=pytz.utc),
                    comment, file_)

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and other.__dict__ == self.__dict__

    def __str__(self):
        return 'Post({0}, {1}, {2})'.format(self.id,
                                            str(self.timestamp),
                                            self.file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division
import sys
import os
import argparse
import logging

import chandl
from chandl import util
from chandl.downloader import Downloader
from chandl.model.thread import Thread
from chandl.model import file


logger = logging.getLogger(__name__)


def _print_error(msg):
    """
    Print a string to stderr.

    :param msg: The message to print.
    """
    print(msg, file=sys.stderr)


def _construct_parser():
    parser = argparse.ArgumentParser(prog='chandl',
                                     description='A lightweight tool for '
                                                 'parsing and downloading '
                                                 '4chan threads.')
    parser.add_argument('-V', '--version',
                        action='version',
                        version='%(prog)s ' + chandl.__version__)
    parser.add_argument('-v', '--verbosity',
                        help='increase output verbosity',
                        action='count',
                        default=0)
    parser.add_argument('-f', '--filter',
                        help='file types or extensions to download, value '
                             'either comma-separated or option passed multiple '
                             'times',
                        action='append',
                        type=util.decode_cli_arg,
                        nargs='?')
    parser.add_argument('-e', '--exclude',
                        help='file names to exclude, value either '
                             'comma-separated or option passed multiple times',
                        action='append',
                        type=util.decode_cli_arg,
                        default=[],
                        nargs='?')
    parser.add_argument('-o', '--output-dir',
                        help='the directory to create the `thread-dir` within',
                        nargs='?',
                        type=util.decode_cli_arg,
                        default=os.getcwd())
    parser.add_argument('-t', '--thread-dir',
                        help='relative to the `output-dir`, this will contain '
                             'downloaded files',
                        type=util.decode_cli_arg,
                        nargs='?')
    parser.add_argument('-n', '--name',
                        help='the format to use for downloaded file names',
                        nargs='?',
                        type=util.decode_cli_arg,
                        default='{file.id} - {file.name}.{file.extension}')
    parser.add_argument('-p', '--parallelism',
                        help='the maximum number of download threads to use '
                             'per core',
                        type=int,
                        default=2)
    parser.add_argument('url',
                        type=util.decode_cli_arg,
                        help='the URL of the thread to download')
    return parser


def _remove_unwanted(posts, args):

    # filter out posts without a file
    posts = [post for post in posts if post.has_file]
    logger.debug('%d contain a file', len(posts))

    # filter out files of the wrong type
    if args.filter:
        extensions = file.expand_filters(args.filter)
        posts = [post for post in posts if post.file.extension in extensions]
    logger.debug('%d are also of the desired format', len(posts))

    # filter out excluded file names
    filenames = util.expand_cli_args(args.exclude)
    posts = [post for post in posts if post.file.name not in filenames]
    logger.debug('%d have also not been excluded', len(posts))

    return posts


def main():
    """
    chandl's command-line entry point.
    """

    parser = _construct_parser()
    args = parser.parse_args()

    # sort out logging output and level
    level = util.log_level_from_vebosity(args.verbosity)
    root = logging.getLogger()
    root.setLevel(level)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter('%(levelname)s %(message)s'))
    root.addHandler(handler)

    logger.debug(args)

    try:
        thread = Thread.from_url(args.url)
    except (ValueError, IOError) as e:
        _print_error('Error retrieving thread: {0}'.format(e))
        return 1

    posts = thread.posts
    logger.debug('Thread contains %d posts', len(posts))

    posts = _remove_unwanted(posts, args)
    logger.debug('Will download %d posts', len(posts))

    # check whether we still have anything to do
    if not posts:
        print('All files are either filtered out or excluded')
        return 0

    # use the first post to validate the --name
    try:
        post = posts[0]
        args.name.format(**post.__dict__)
    except KeyError as e:
        _print_error('Invalid file name specifier: {0}'.format(e))
        return 2

    # set an appropriate thread_dir if one was not specified
    if not args.thread_dir:
        args.thread_dir = util.make_filename(thread.title)

    # create --thread-dir
    write_dir = os.path.join(args.output_dir, args.thread_dir)
    if not os.path.isdir(write_dir):
        try:
            os.mkdir(write_dir, 0o700)
        except OSError as e:
            _print_error(
                'Failed to create the thread directory at {0}: {1}'.format(
                    write_dir, e))
            return 3

    # download the files
    print('Saving \'{0}\' to \'{1}\''.format(thread.title,
                                             os.path.relpath(write_dir,
                                                             os.getcwd())))
    downloader = Downloader(write_dir, args.name, args.parallelism)
    print(downloader.download(posts, level >= logging.WARNING))

    return 0


if __name__ == '__main__':
    status = main()
    logger.debug('Exiting with status %d', status)
    sys.exit(status)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import unittest
import os
from httmock import all_requests, response, HTTMock
from pyfakefs import fake_filesystem_unittest

from chandl import util
from chandl.model import file
from chandl.model.file import File

from chandl.tests.model.test_post import TestPost

_REAL_OPEN = open


class TestExpandFilters(unittest.TestCase):

    def test_empty(self):
        self.assertEqual(file.expand_filters([]), set())

    def test_duplicates(self):
        self.assertEqual(file.expand_filters(['webm', 'jpg', 'webm']),
                         {'webm', 'jpg'})

    def test_expand_type(self):
        self.assertEqual(file.expand_filters(file._FILTERS['videos']),
                         set(file.TYPE_VIDEO))

    def test_csv(self):
        self.assertEqual(file.expand_filters(['webm,jpg,webm', 'png,jpg']),
                         {'webm', 'jpg', 'png'})

    def test_strip(self):
        self.assertEqual(file.expand_filters(['webm  ', ' jpg', '   webm']),
                         {'webm', 'jpg'})

    def test_mixed(self):
        self.assertEqual(file.expand_filters(['webm , gif ', ' jpg,png',
                                              '   gif,webm,videos']),
                         set(['webm', 'gif', 'jpg', 'png'] + file.TYPE_VIDEO))


class TestFile(fake_filesystem_unittest.TestCase):

    _RESOURCES_DIR = os.path.join(os.path.dirname(__file__), 'resources')

    @classmethod
    def setUpClass(cls):
        cls.file = File(TestPost.POST_JSON['tim'], TestPost.BOARD,
                        TestPost.POST_JSON['filename'],
                        TestPost.POST_JSON['ext'][1:],
                        TestPost.POST_JSON['fsize'], TestPost.POST_JSON['w'],
                        TestPost.POST_JSON['h'], TestPost.POST_JSON_FILE_MD5)

    def setUp(self):
        self.setUpPyfakefs()

    def test_id(self):
        self.assertEqual(self.file.id, TestPost.POST_JSON['tim'])

    def test_board(self):
        self.assertEqual(self.file.board, TestPost.BOARD)

    def test_name(self):
        self.assertEqual(self.file.name, TestPost.POST_JSON['filename'])

    def test_extension(self):
        self.assertEqual(self.file.extension, TestPost.POST_JSON['ext'][1:])

    def test_size(self):
        self.assertEqual(self.file.size, TestPost.POST_JSON['fsize'])

    def test_width(self):
        self.assertEqual(self.file.width, TestPost.POST_JSON['w'])

    def test_height(self):
        self.assertEqual(self.file.height, TestPost.POST_JSON['h'])

    def test_md5(self):
        self.assertEqual(self.file.md5, TestPost.POST_JSON_FILE_MD5)

    def test_parse_json_no_file(self):
        with self.assertRaises(ValueError):
            File.parse_json(TestPost.BOARD, TestPost.POST_NO_FILE_JSON)

    def test_parse_json_file(self):
        self.assertEqual(File.parse_json(TestPost.BOARD, TestPost.POST_JSON),
                         self.file)

    def test_url(self):
        self.assertEqual(self.file.url, TestPost.POST_JSON_FILE_URL)

    def test_save_to_exists(self):
        # to ensure it's not re-downloading, make any re-download fail by not
        # configuring HTTMock

        directory = '/tmp'
        name = 'dl.jpg'
        path = os.path.join(directory, name)

        # copy real file to fakefs
        with _REAL_OPEN(
                os.path.join(self._RESOURCES_DIR,
                             TestPost.POST.file.filename), 'rb') as real_file:
            self.fs.CreateFile(path, contents=real_file.read())

        self.file.save_to(directory, name)

    def test_save_to_exists_md5_mismatch(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            with _REAL_OPEN(os.path.join(
                    self._RESOURCES_DIR,
                    TestPost.POST.file.filename), 'rb') as f:
                return response(content=f.read(), stream=True)

        directory = '/tmp'
        name = 'corrupt.jpg'
        path = os.path.join(directory, name)

        self.fs.CreateFile(path, contents='corrupt content')
        with HTTMock(response_content):
            self.file.save_to(directory, name)
        self.assertEqual(util.md5_file(path), TestPost.POST.file.md5)

    def test_save_to_non_200(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            return response(404)

        with HTTMock(response_content), self.assertRaises(IOError):
            self.file.save_to(self._RESOURCES_DIR, self.file.filename)

    def test_save_to_verify_mismatch(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            return response(content='corrupt content', stream=True)

        with HTTMock(response_content), self.assertRaises(IOError):
            self.fs.CreateDirectory(self._RESOURCES_DIR)
            self.file.save_to(self._RESOURCES_DIR, 'dl.jpg')

    def test_save_to_ok(self):
        # noinspection PyUnusedLocal
        @all_requests
        def response_content(url, request):
            with _REAL_OPEN(os.path.join(
                    self._RESOURCES_DIR,
                    TestPost.POST.file.filename), 'rb') as f:
                return response(content=f.read(), stream=True)

        with HTTMock(response_content):
            self.fs.CreateDirectory(self._RESOURCES_DIR)
            self.file.save_to(self._RESOURCES_DIR, 'dl.jpg')

        self.assertEqual(
            util.md5_file(os.path.join(self._RESOURCES_DIR,
                                       'dl.jpg')),
            TestPost.POST.file.md5)

    def test_str(self):
        self.assertEqual(str(self.file),
                         'File({0}, {1}.{2}, {3}, {4}x{5})'.format(
                             TestPost.POST_JSON['tim'],
                             TestPost.POST_JSON['filename'],
                             TestPost.POST_JSON['ext'][1:],
                             util.bytes_fmt(TestPost.POST_JSON['fsize']),
                             TestPost.POST_JSON['w'],
                             TestPost.POST_JSON['h']))
